local bodyName = "body"
local shapeName = "shape"

local bgTexture = Texture.new("bg.png", true, {wrap = Texture.REPEAT})
local bg = Shape.new()
bg:setFillStyle(Shape.TEXTURE, bgTexture)
bg:beginPath()
bg:moveTo(0,0)
bg:lineTo(application:getDeviceWidth(), 0)
bg:lineTo(application:getDeviceWidth(), application:getDeviceHeight())
bg:lineTo(0, application:getDeviceHeight())
bg:closePath()
bg:endPath()

local img = Bitmap.new(Texture.new("star.png"))
img:setPosition((application:getDeviceWidth() / 2) - (img:getWidth() / 2), 16)

local anchorPoint = Bitmap.new(Texture.new("anchor_point.png"))
anchorPoint:setAlpha(0.75)
anchorPoint:setAnchorPoint(0.49, 0.49)
anchorPoint:setPosition(img:getX(), img:getY())

local nodes = Sprite.new()
local nodeTexture = Texture.new("node.png")
local nodeFocusTexture = Texture.new("node_focus.png")

local polygonContour = Shape.new()
function polygonContour:draw()
	self:clear()
	if nodes:getNumChildren() > 1 then
		self:setLineStyle(2, 0x0094ff, 1)
		self:setFillStyle(Shape.SOLID, 0x0094ff, 0.25)
		self:beginPath()
		self:moveTo(nodes:getChildAt(1):getX(), nodes:getChildAt(1):getY())
		for i = 2, nodes:getNumChildren() do
			self:lineTo(nodes:getChildAt(i):getX(), nodes:getChildAt(i):getY())
		end
		self:closePath()
		self:endPath()
	end
end

local trianglesContour = Shape.new()

local imgBorder = Shape.new()
imgBorder:setLineStyle(1, 0x000000, 1)
imgBorder:beginPath()
imgBorder:moveTo(img:getX(), img:getY())
imgBorder:lineTo(img:getX() + img:getWidth(), img:getY())
imgBorder:lineTo(img:getX() + img:getWidth(), img:getY() + img:getHeight())
imgBorder:lineTo(img:getX(), img:getY() + img:getHeight())
imgBorder:closePath()
imgBorder:endPath()

local btnUp = Bitmap.new(Texture.new("btn_up.png"))
local btnDown = Bitmap.new(Texture.new("btn_down.png"))
local btn = Sprite.new()
btn.focus = false
btn:addChild(btnUp)
btn:setPosition((application:getDeviceWidth() / 2) - (btn:getWidth() / 2), img:getY() + img:getHeight() + 16)

function inTriangle(p, v1, v2, v3)
	local a = (v1.x - p.x) * (v2.y - v1.y) - (v2.x - v1.x) * (v1.y - p.y)
	local b = (v2.x - p.x) * (v3.y - v2.y) - (v3.x - v2.x) * (v2.y - p.y)
	local c = (v3.x - p.x) * (v1.y - v3.y) - (v1.x - v3.x) * (v3.y - p.y)
	return (a >= 0 and b >= 0 and c >= 0) or (a <= 0 and b <= 0 and c <= 0)
end

function inPolygon(p, v)
	local c = 0
	local x0 = v[#v].x - p.x
	local y0 = v[#v].y - p.y
	for i = 1, #v do
		local x1 = v[i].x - p.x
		local y1 = v[i].y - p.y
		if (y0 > 0 and y1 <= 0 and x1 * y0 > y1 * x0) or (y1 > 0 and y0 <= 0 and x0 * y1 > y0 * x1) then
			c = c + 1
		end
		x0 = x1
		y0 = y1
	end
	return (c % 2) == 1
end

function ckw(v)
	local sum = 0
	for i = 2, #v do
		sum = sum + (v[i].x - v[i - 1].x) * (v[i].y + v[i - 1].y)
	end
	sum = sum + (v[1].x - v[#v].x) * (v[1].y + v[#v].y)
	if sum > 0 then
		for i = 1, #v / 2 do
			v[i], v[#v + 1 - i] =  v[#v + 1 - i], v[i]
		end
	end
	return v
end

function getStarted()
	print("-------------------------------------------------------")
	print("local anchorPoint = {x = " .. anchorPoint:getX() .. ", y = " .. anchorPoint:getY() .. "}")
	print("local polygon = {")
	local v = {}
	local triangles = {}
	for i = 1, nodes:getNumChildren() do
		v[#v + 1] = {x = nodes:getChildAt(i):getX(), y = nodes:getChildAt(i):getY()}
		print("	{x = " .. v[#v].x - anchorPoint:getX() .. 
			", y = " .. v[#v].y - anchorPoint:getY() .. "},")
	end
	print("}")
	trianglesContour:clear()
	trianglesContour:setLineStyle(2, 0xff9400, 1)
	trianglesContour:beginPath()
	local v1, v2, v3 = 1, 2, 3
	while #v > 3 do
		local success = true
		for i = 1, #v do
			if i ~= v1 and i ~= v2 and i ~= v3 then
				if inTriangle(v[i], v[v1], v[v2], v[v3]) then
					success = false
					do break end
				end
			end
		end
		if not inPolygon({x = (v[v1].x + v[v2].x + v[v3].x) / 3, y = (v[v1].y + v[v2].y + v[v3].y) / 3}, v) then
			success = false
		end
		if success then
			trianglesContour:moveTo(v[v1].x - 1, v[v1].y - 1)
			trianglesContour:lineTo(v[v3].x - 1, v[v3].y - 1)
			triangles[#triangles + 1] = ckw({v[v1], v[v2], v[v3]})
			table.remove(v, v2)
		end
		v1 = v1 + 1
		if v1 > #v then v1 = 1 end
		v2 = v1 + 1
		if v2 > #v then v2 = 1 end
		v3 = v2 + 1
		if v3 > #v then v3 = 1 end
	end
	trianglesContour:endPath()
	triangles[#triangles + 1] = ckw(v)
	
	print("local triangles = {")
	for i = 1, #triangles do
		print("\t{" ..
			"{x = " .. triangles[i][1].x - anchorPoint:getX() .. ", y = " .. triangles[i][1].y - anchorPoint:getY() .. "}, " .. 
			"{x = " .. triangles[i][2].x - anchorPoint:getX() .. ", y = " .. triangles[i][2].y - anchorPoint:getY() .. "}, " ..
			"{x = " .. triangles[i][3].x - anchorPoint:getX() .. ", y = " .. triangles[i][3].y - anchorPoint:getY() .. "}}, ")
	end
	print("}")
	print("local shape = b2.PolygonShape.new()")
	print("for i = 1, #triangles do")
	print("	shape:set(triangles[i][1].x, triangles[i][1].y, " .. 
		"triangles[i][2].x, triangles[i][2].y, " ..
		"triangles[i][3].x, triangles[i][3].y)")
	print("	body:createFixture({shape = shape})")
	print("end")
	
end

local t0 = os.time()
function onMouseDown(e)
	if btn:hitTestPoint(e.x, e.y) then
		btn.focus = true
		btn:removeChild(btnUp)
		btn:addChild(btnDown)
		return
	end
	if anchorPoint:hitTestPoint(e.x, e.y) then
		anchorPoint.x0 = e.x
		anchorPoint.y0 = e.y
		anchorPoint.isDragging = true
		return
	end
	for i = 1, nodes:getNumChildren() do
		if nodes:getChildAt(i):hitTestPoint(e.x, e.y) then
			if (os.time() - t0) == 0 and nodes:getChildAt(i).isFocus then
				trianglesContour:clear()
				nodes:removeChildAt(i)
				if nodes:getNumChildren() == 0 then
					return
				elseif i == 1 then
					nodes:getChildAt(nodes:getNumChildren()).isFocus = true
					nodes:getChildAt(nodes:getNumChildren()):setTexture(nodeFocusTexture)
				else
					nodes:getChildAt(i - 1).isFocus = true
					nodes:getChildAt(i - 1):setTexture(nodeFocusTexture)
				end
				polygonContour:draw()
				return
			end
			for j = 1, nodes:getNumChildren() do
				nodes:getChildAt(j).isFocus = false
				nodes:getChildAt(j):setTexture(nodeTexture)
			end
			nodes:getChildAt(i).x0 = e.x
			nodes:getChildAt(i).y0 = e.y
			nodes:getChildAt(i).isFocus = true
			nodes:getChildAt(i).isDragging = true
			nodes:getChildAt(i):setTexture(nodeFocusTexture)
			t0 = os.time()
			return
		end
	end
	if img:hitTestPoint(e.x, e.y) then
		trianglesContour:clear()
		local pos = 1
		if nodes:getNumChildren() > 0 then
			for i = 1, nodes:getNumChildren() do
				if nodes:getChildAt(i).isFocus then
					nodes:getChildAt(i).isFocus = false
					nodes:getChildAt(i):setTexture(nodeTexture)
					pos = i + 1
					do break end
				end
			end
		end
		nodes:addChildAt(Bitmap.new(nodeFocusTexture), pos)
		nodes:getChildAt(pos).x0 = e.x
		nodes:getChildAt(pos).y0 = e.y
		nodes:getChildAt(pos).isFocus = true
		nodes:getChildAt(pos).isDragging = true
		nodes:getChildAt(pos):setAnchorPoint(0.49, 0.49)
		nodes:getChildAt(pos):setPosition(e.x, e.y)
		polygonContour:draw()
		return
	end
end

function dragging(obj, e)
	local dx = e.x - obj.x0
	local dy = e.y - obj.y0
	obj:setX(obj:getX() + dx)
	obj:setY(obj:getY() + dy)
	if e.x < img:getX() then
		obj:setX(img:getX())
	end
	if e.x > (img:getX() + img:getWidth()) then
		obj:setX(img:getX() + img:getWidth())
	end
	if e.y < img:getY() then
		obj:setY(img:getY())
	end
	if e.y > (img:getY() + img:getHeight()) then
		obj:setY(img:getY() + img:getHeight())
	end
	obj.x0 = e.x
	obj.y0 = e.y
end

function onMouseMove(e)
	if anchorPoint.isDragging then
		dragging(anchorPoint, e)
		return
	end
	for i = 1, nodes:getNumChildren() do
		if nodes:getChildAt(i).isDragging then
			trianglesContour:clear()
			dragging(nodes:getChildAt(i), e)
			polygonContour:draw()
			return
		end
	end
end

function onMouseUp(e)
	if btn.focus then
		if btn:hitTestPoint(e.x, e.y) then
			getStarted()
		end
		btn.focus = false
		btn:removeChild(btnDown)
		btn:addChild(btnUp)
		return
	end
	if anchorPoint.isDragging then
		anchorPoint.isDragging = false
		return
	end
	for i = 1, nodes:getNumChildren() do
		if nodes:getChildAt(i).isDragging then
			nodes:getChildAt(i).isDragging = false
			return
		end
	end
end

stage:addEventListener(Event.MOUSE_DOWN, onMouseDown)
stage:addEventListener(Event.MOUSE_MOVE, onMouseMove)
stage:addEventListener(Event.MOUSE_UP, onMouseUp)

stage:addChild(bg)
stage:addChild(img)
stage:addChild(polygonContour)
stage:addChild(trianglesContour)
stage:addChild(nodes)
stage:addChild(anchorPoint)
stage:addChild(imgBorder)
stage:addChild(btn)
